from canvasapi.assignment import Assignment
from canvasapi.file import File
from canvasapi.page import Page
from canvasapi.paginated_list import PaginatedList
from canvasapi.course import Course
from datetime import date

class CanvasClientGetters():
    def __init__(self, canvas, course: Course):
        self._canvas = canvas
        self._course = course

    def get_announcements(self) -> PaginatedList:
        return self._canvas.get_announcements(context_codes=[self._course],
                                                            start_date="2016-01-01",
                                                            end_date=date.today().isoformat())
    
    def get_assignments(self) -> PaginatedList:
        return self._course.get_assignments()
    
    def get_assignment(self, assignment_id) -> Assignment:
        return self._course.get_assignment(assignment_id)
    
    def get_files(self) -> PaginatedList:
        return self._course.get_files()
    
    def get_file(self, file_id) -> File:
        return self._course.get_file(file_id)

    def get_modules(self) -> PaginatedList:
        return self._course.get_modules()
    
    def get_pages(self) -> PaginatedList:
        return self._course.get_pages(published=True,
                                     include=['body'])
    
    def get_page(self, url) -> Page:
        return self._course.get_page(url)
    
    def get_syllabus(self) -> str:
        return self._course.syllabus_body



import logging

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)
# Quiet noisy libraries
logging.getLogger("canvasapi").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)
logging.getLogger("LangChainKaltura").setLevel(logging.WARNING)

class Logger():
    def __init__(self):
        self.progress = []
        self.errors = []

    def logStatement(self, message: str, level: str):
        """Log messages and track progress"""
        match level:
            case "INFO":
                logger.info(message)
            case "DEBUG":
                logger.debug(message)
            case "WARNING":
                logger.warning(message)
                self.errors.append({"message": message, "level": level})
        self.progress.append({"message": message, "level": level})


    def _filtered_statements_by_level(self, level: str) -> list:
        """Returns statements corresponding to desired output level"""
        return [statement for statement in self.progress if statement.level == level]



from urllib.parse import urljoin
from canvasapi.exceptions import CanvasException
from langchain.docstore.document import Document

from canvas_langchain.base import BaseSectionLoader, BaseSectionLoaderVars
from canvasapi.page import Page

class PageLoader(BaseSectionLoader):
    def __init__(self, baseSectionVars: BaseSectionLoaderVars, course_api: str):
        super().__init__(baseSectionVars)
        self.course_api = course_api

    def load_section(self) -> list[Document]:
        self.logger.logStatement(message='Loading pages...\n', level="INFO")
        page_documents = []

        try:
            pages = self.canvas_client_extractor.get_pages()
            for page in pages:
                page_documents.extend(self._load_item(page))

        except CanvasException:
            self.logger.logStatement(message=f"Canvas exception loading pages", level="WARNING")
        
        return page_documents

    def _load_item(self, page: Page) -> list[Document]:
        """Loads and formats a single page and its embedded URL(s) content """
        try:
            if not page.locked_for_user and page.body and f"Page:{page.page_id}" not in self.indexed_items:
                self.logger.logStatement(message=f"Loading page: {page.title}", level="DEBUG")
                self.indexed_items.add(f"Page:{page.page_id}")                      

                page_body = self.parse_html(html=page.body)
            
                page_url = urljoin(self.course_api, f'pages/{page.url}')
                metadata={"content": page_body,
                        "data": {"filename": page.title,
                                "source": page_url,
                                "kind": "page",
                                "id": page.page_id}
                        }
                return self.process_data(metadata=metadata)
        except Exception as error:
            self.logger.logStatement(message=f"Error loading page {page.title}: {error}",
                                     level="WARNING")
        return []

    def load_from_module(self, item: Page, **kwargs) -> list[Document]:
        """Loads page from module item"""
        self.logger.logStatement(message=f"Loading page {item.page_url} from module.", 
                                    level="DEBUG")
        page = self.canvas_client_extractor.get_page(url=item.page_url)
        return self._load_item(page)



from canvasapi import Canvas
from canvasapi.exceptions import Forbidden
from canvasapi.course import Course
from urllib.parse import urljoin

from canvas_langchain.utils.logging import Logger
from canvas_langchain.sections.announcements import AnnouncementLoader
from canvas_langchain.sections.assignments import AssignmentLoader
from canvas_langchain.sections.files import FileLoader
from canvas_langchain.sections.modules import ModuleLoader
from canvas_langchain.sections.pages import PageLoader
from canvas_langchain.sections.syllabus import SyllabusLoader
from canvas_langchain.base import BaseSectionLoaderVars, BaseSectionLoader

from canvas_langchain.client_getters import CanvasClientGetters

class UnpublishedCourseException(Exception):
    def __init__(self, message):
        super().__init__(message)

class CanvasClient():
    def __init__(self, api_url: str, api_key: str, course_id: int):
        self._canvas = Canvas(api_url, api_key)
        self.api_url = api_url
        self._course = self.get_course(course_id)
        self.content_extractor = CanvasClientGetters(self._canvas, self._course)

    def get_course(self, course_id: int) -> Course:
        try:
            return self._canvas.get_course(course_id, include=[ "syllabus_body" ])
        except Forbidden:
            exception_message = (
                "User forbidden from accessing Canvas course. " 
                "Please check user permissions and course availability."
            )
            raise UnpublishedCourseException(message=exception_message)

    def get_available_tabs(self) -> list[str]:
        return [tab.label for tab in self._course.get_tabs()]

    def get_loaders(self, index_external_urls: bool, logger: Logger) -> dict[str, BaseSectionLoader]:
        base_vars = BaseSectionLoaderVars(canvas_client_extractor=self.content_extractor, indexed_items=set(), logger=logger)
        course_api = urljoin(self.api_url, f'courses/{self._course.id}/')
        
        assignment_loader = AssignmentLoader(baseSectionVars=base_vars)
        page_loader = PageLoader(baseSectionVars=base_vars, course_api=course_api)
        file_loader = FileLoader(baseSectionVars=base_vars, course_api=course_api)

        return {
            "Announcements": AnnouncementLoader(baseSectionVars=base_vars),
            "Assignments": assignment_loader,
            "Files": file_loader,
            "Modules": ModuleLoader(baseSectionVars=base_vars, index_external_urls=index_external_urls, 
                                    loaders={
                                        "Pages": page_loader,
                                        "Assignments": assignment_loader,
                                        "Files": file_loader
                                    }),
            "Pages": page_loader,
            "Syllabus": SyllabusLoader(baseSectionVars=base_vars, course_api=course_api)
        }



from urllib.parse import urljoin

from canvas_langchain.base import BaseSectionLoader, BaseSectionLoaderVars
from langchain.docstore.document import Document

class SyllabusLoader(BaseSectionLoader):
    def __init__(self, baseSectionVars: BaseSectionLoaderVars, course_api: str):
        super().__init__(baseSectionVars)
        self.course_api = course_api

    def load_section(self) -> list[Document]:
        self.logger.logStatement(message='Loading syllabus...\n', level="INFO")
        try:
            syllabus_body = self.canvas_client_extractor.get_syllabus()
            if syllabus_body:
                syllabus_text = self.parse_html(syllabus_body)
                syllabus_url = urljoin(self.course_api, 'assignments/syllabus')

                metadata={"content": syllabus_text,
                        "data": {"filename": "Course Syllabus",
                                "source": syllabus_url,
                                "kind": "syllabus"}
                            }
                return self.process_data(metadata=metadata)

        except AttributeError as err:
            self.logger.logStatement(message=f"Attribute error loading syllabus: {err}", level="WARNING")

        except Exception as err:
            self.logger.logStatement(message=f"Error loading syllabus: {err}", level="WARNING")

        return []



from datetime import datetime, timezone
from canvasapi.exceptions import CanvasException
from canvasapi.module import ModuleItem
from langchain_community.document_loaders import UnstructuredURLLoader
from canvas_langchain.base import BaseSectionLoader, BaseSectionLoaderVars
from langchain.docstore.document import Document

class ModuleLoader(BaseSectionLoader):
    def __init__(self, baseSectionVars: BaseSectionLoaderVars, index_external_urls: bool, loaders: dict[str, BaseSectionLoader]):
        super().__init__(baseSectionVars)
        self.loaders = loaders
        self.index_external_urls = index_external_urls

    def load_section(self) -> list[Document]:
        """Loads content from all unlocked modules in course"""
        self.logger.logStatement(message='Loading modules...\n', level="INFO")
        module_documents = []
        try:
            modules = self.canvas_client_extractor.get_modules()
            for module in modules:
                module_documents.extend(self._load_item(module))

        except CanvasException as ex:
            self.logger.logStatement(message=f"Canvas exception loading modules. Error: {ex}", level="WARNING")

        return module_documents
    
    def _load_item(self, module: ModuleItem) -> list[Document]:
        """Loads content from a single module"""
        locked, formatted_datetime = self._get_module_metadata(module.unlock_at)
        module_items = module.get_module_items()
        module_docs = []
        try:
            for item in module_items:
                if item.type in ["Page", "File", "Assignment"]:
                    module_docs.extend(self.loaders[f"{item.type}s"].load_from_module(item=item,
                                                                                      module_name=module.name,
                                                                                      locked=locked,
                                                                                      formatted_datetime=formatted_datetime
                                                                                      ))
                elif item.type == "ExternalUrl" and self.index_external_urls:
                    module_docs.extend(self._load_external_url(item))
        except CanvasException as ex:
            self.logger.logStatement(message=f"Canvas exception loading module items. Error: {ex}", level="WARNING")
        except Exception as ex:
            self.logger.logStatement(message=f"Error loading module items. Error: {ex}", level="WARNING")
        
        return module_docs

    def _get_module_metadata(self, unlock_time: str) -> tuple[bool, datetime | str]:
        """Returns if module is locked and corresponding unlock time ("" if unlocked)"""
        locked=False
        formatted_datetime=""
        if unlock_time:
            # get formatted unlock time
            formatted_datetime = datetime.strptime(unlock_time, "%Y-%m-%dT%H:%M:%SZ").replace(tzinfo=timezone.utc)
            current_time = datetime.now(timezone.utc)
            # determine if locked
            locked = current_time < formatted_datetime

        return locked, formatted_datetime

    def _load_external_url(self, item: ModuleItem) -> list[Document]:
        """Loads external URL from module item"""
        if item.external_url and f"ExtUrl:{item.external_url}" not in self.indexed_items:
            self.logger.logStatement(message=f"Loading external url {item.external_url} from module.", 
                                    level="DEBUG")
            url_loader = UnstructuredURLLoader(urls = [item.external_url])
            docs = url_loader.load()
            if docs:
                self.indexed_items.add(f"ExtUrl:{item.external_url}")
                return docs
        return []



import tempfile
from io import BytesIO
from PyPDF2 import PdfReader
from PyPDF2 import errors
from urllib.parse import urljoin
from canvas_langchain.base import BaseSectionLoaderVars
from canvasapi.file import File

from canvas_langchain.base import BaseSectionLoader
from canvasapi.exceptions import CanvasException, ResourceDoesNotExist

from langchain.docstore.document import Document

from langchain_community.document_loaders import (
    Docx2txtLoader,
    UnstructuredExcelLoader,
    UnstructuredMarkdownLoader,
    UnstructuredPowerPointLoader,
    CSVLoader
) 

class FileLoader(BaseSectionLoader):
    def __init__(self, baseSectionVars: BaseSectionLoaderVars, course_api: str):
        super().__init__(baseSectionVars)
        self.course_api = course_api
        self.type_match = {
            "text/md": "md",
            "text/csv": "csv",
            "application/vnd.ms-excel": "excel",
            "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet" : "excel",
            "application/vnd.openxmlformats-officedocument.wordprocessingml.document": "docx",
            "application/vnd.openxmlformats-officedocument.presentationml.presentation": "pptx"
        }

    def load_section(self) -> list[Document]:
        """Loads and formats all files from Canvas course"""
        self.logger.logStatement(message='Loading files...\n', level="INFO")

        file_documents = []
        try:
            files = self.canvas_client_extractor.get_files()
            for file in files:
                file_documents.extend(self._load_item(file))

        except CanvasException as error:
            self.logger.logStatement(message=f"Canvas exception loading files {error}",
                                     level="WARNING")
        return file_documents
    
    def _load_item(self, file: File) -> list[Document]:
        """Loads given file based on extension"""
        if f"File:{file.id}" not in self.indexed_items:
            self.indexed_items.add(f"File:{file.id}")
            try:
                content_type = getattr(file, "content-type")
                self.logger.logStatement(message=f"Loading file: {file.filename}", level="DEBUG")

                if content_type in ["text/plain", "text/rtf"]:
                    return self._load_rtf_or_text_file(file)
                elif content_type == "text/html":
                    return self._load_html_file(file)
                elif content_type == "application/pdf":
                    return self._load_pdf_file(file)
                elif content_type in self.type_match:
                    return self._load_file_general(file, self.type_match[content_type])
            
            # exception occurs when file is in a hidden module
            except ResourceDoesNotExist as err:
                self.logger.logStatement(message=f"File {file.filename} does not exist - likely in hidden module {err}",
                                        level="DEBUG")

            except Exception as ex:
                self.logger.logStatement(message=f"Error loading file {file.filename}: {ex}",
                        level="DEBUG")

        return []

    def load_from_module(self, item: File, **kwargs):
        """Loads file from module item"""
        self.logger.logStatement(message=f"Loading file {item.content_id} from module.", 
                                 level="DEBUG")
        file = self.canvas_client_extractor.get_file(file_id=item.content_id)
        return self._load_item(file)

    def _load_rtf_or_text_file(self, file: File) -> list[Document]:
        """Loads and formats text and rtf file data"""
        file_contents = file.get_contents(binary=False)
        metadata={'content': file_contents,
                  "data": {"filename": file.filename, 
                           "source": file.url, 
                           "kind": "file",
                           "id": file.id }

        }
        return self.process_data(metadata=metadata)

    def _load_html_file(self, file: File) -> list[Document]:
        """Loads and formats html file data"""
        file_contents = file.get_contents(binary=False)
        file_text = self.parse_html(html=file_contents)
        metadata={"content":file_text,
                "data": {"name": file.filename,
                        "source":file.url,
                        "kind":"file",
                        "id": file.id}
                    }
        return self.process_data(metadata=metadata)

    def _load_pdf_file(self, file: File) -> list[Document]:
        """Loads given pdf file by page"""
        file_contents = file.get_contents(binary=True)
        docs = []
        try:
            pdf_reader = PdfReader(BytesIO(file_contents))
            # extract info by page
            for i, page in enumerate(pdf_reader.pages):
                metadata={"content": page.extract_text(),
                          "data": {"filename": file.filename,
                                              "source": urljoin(self.course_api, f"files/{file.id}"),
                                              "kind": "file",
                                              "page": i+1}

                        }       
                docs.extend(self.process_data(metadata=metadata))
        except errors.FileNotDecryptedError:
            self.logger.logStatement(message=f"Error: pdf {file.filename} is encrypted.",
                                     level="WARNING")
        except Exception as err:
            self.logger.logStatement(message=f"Error loading pdf {file.filename}. Err: {err}",
                                     level="WARNING")
        return docs

    def _load_file_general(self, file: File, file_type: str) -> list[Document]:
        """Loads docx, excel, pptx, and md files"""
        file_contents = file.get_contents(binary=True)
        docs=[]
        try:
            with tempfile.TemporaryDirectory() as temp_dir:
                file_path = f"{temp_dir}/{file.filename}"

                with open(file_path, "wb") as binary_file:
                    # Write bytes to file
                    binary_file.write(file_contents)

                match file_type:
                    case 'csv':
                        loader = CSVLoader(file_path)
                    case 'excel':
                        loader = UnstructuredExcelLoader(file_path)         
                    case 'docx':
                        loader=Docx2txtLoader(file_path)
                    case 'md':
                        loader = UnstructuredMarkdownLoader(file_path)
                    case 'pptx':
                        loader=UnstructuredPowerPointLoader(file_path)
                if loader:
                    docs = loader.load()
                    for i, _ in enumerate(docs):
                        docs[i].page_content = self._remove_null_bytes(docs[i].page_content)
                        docs[i].metadata["filename"] = file.filename
                        docs[i].metadata["source"] = urljoin(self.course_api, f"files/{file.id}")
        except Exception:
            self.logger.logStatement(message=f"Error loading {file.filename}", level="WARNING")
        return docs



from dataclasses import dataclass
from bs4 import BeautifulSoup
from abc import ABC, abstractmethod

from langchain.docstore.document import Document
from canvas_langchain.utils.logging import Logger

from canvasapi.discussion_topic import DiscussionTopic
from canvasapi.assignment import Assignment
from canvasapi.file import File
from canvasapi.module import ModuleItem
from canvasapi.page import Page
from canvas_langchain.client_getters import CanvasClientGetters


@dataclass
class BaseSectionLoaderVars:
    canvas_client_extractor: CanvasClientGetters
    indexed_items: set
    logger: Logger


class BaseSectionLoader(ABC):
    """Abstract base class for loading sections of a Canvas course"""
    def __init__(self, baseSectionVars: BaseSectionLoaderVars):
        self.canvas_client_extractor = baseSectionVars.canvas_client_extractor
        self.indexed_items = baseSectionVars.indexed_items
        self.logger = baseSectionVars.logger

    @abstractmethod
    def load_section(self) -> list[Document]:
        """Load section data and return a list of Document objects"""
        pass

    def _load_item(self, item: File | Assignment | Page | DiscussionTopic | ModuleItem) -> list[Document]:
        """Load a single section item and return a list of Document objects"""
        raise NotImplementedError("This optional method should be implemented in subclass")

    def load_from_module(self, item: File | Assignment | Page, 
                         module_name: str | None = None, 
                         locked: bool | None = None, 
                         formatted_datetime: str | None = None) -> list[Document]:
        """Load a section item from a module"""
        raise NotImplementedError("This optional method should be implemented in subclass")

    def parse_html(self, html: str) -> str:
        """Extracts text and a list of embedded urls from HTML content"""
        bs = BeautifulSoup(html, 'lxml')
        doc_text = bs.text.strip()
        return doc_text
    
    def process_data(self, metadata: dict) -> list[Document]:
        """Process metadata on a single 'page'"""
        document_arr = []    
        if metadata['content']:
            document_arr.append(Document(
                page_content=self._remove_null_bytes(metadata['content']),
                metadata=self._remove_null_bytes(metadata['data'])
            ))
        return document_arr
    
    def _remove_null_bytes(self, metadata_item: str | dict) -> str | dict:
        """Recursively remove NUL bytes from string or dict of strings"""
        if isinstance(metadata_item, str):
            return metadata_item.replace("\x00", '')
        elif isinstance(metadata_item, dict):
            return {key: self._remove_null_bytes(value) for key, value in metadata_item.items()}
        return metadata_item




from canvasapi.assignment import Assignment
from canvasapi.exceptions import CanvasException
from canvas_langchain.base import BaseSectionLoader, BaseSectionLoaderVars
from langchain.docstore.document import Document

class AssignmentLoader(BaseSectionLoader):
    def __init__(self, baseSectionVars: BaseSectionLoaderVars):
        super().__init__(baseSectionVars)

    def load_section(self) -> list[Document]:
        """Load all assignments for a Canvas course"""
        self.logger.logStatement(message='Loading assignments...\n', level="INFO")

        assignment_documents = []
        try:
            assignments = self.canvas_client_extractor.get_assignments()
            for assignment in assignments:
                if f"Assignment:{assignment.id}" not in self.indexed_items:
                    self.indexed_items.add(f"Assignment:{assignment.id}")
                    assignment_documents.extend(self._load_item(assignment, None))

        except CanvasException as error:
            self.logger.logStatement(message=f"Canvas exception loading assignments {error}",
                                    level="WARNING")

        return assignment_documents

    def _load_item(self, assignment: Assignment, description: str | None) -> list[Document]:
        """Load and format one assignment"""
        assignment_description = ""
        self.logger.logStatement(message=f"Loading assignment: {assignment.name}", level="DEBUG")
        try:
            # Custom description from locked module
            if description is not None:
                assignment_description = description

            elif assignment.description:
                assignment_description = self.parse_html(assignment.description)
                                                                
            assignment_content = (
                f"Name: {assignment.name}\n"
                f"Due Date: {assignment.due_at}\n"
                f"Points Possible: {assignment.points_possible}\n"
                f"Description: {assignment_description}\n"
            )

            metadata={"content":assignment_content,
                    "data": {"filename": assignment.name,
                            "source": assignment.html_url,
                            "kind": "assignment",
                            "id": assignment.id}
                        }

            return self.process_data(metadata=metadata)
        except Exception as error:
            self.logger.logStatement(message=f"Error loading assignment {assignment.name}: {error}",
                                     level="WARNING")
        return []

    def load_from_module(self, 
                         item:Assignment, 
                         module_name: str, 
                         locked: bool, 
                         formatted_datetime: str | None) -> list[Document]:
        """Loads assignment from module item"""
        self.logger.logStatement(message=f"Loading assignment {item.content_id} from module.", level="DEBUG")
        assignment = self.canvas_client_extractor.get_assignment(assignment_id=item.content_id)
        description=None
        if locked and formatted_datetime:
            description=f"Assignment is part of module {module_name}, which is locked until {formatted_datetime}"
        return self._load_item(assignment, description)



from canvasapi.exceptions import CanvasException
from langchain.docstore.document import Document
from canvasapi.discussion_topic import DiscussionTopic
from canvas_langchain.base import BaseSectionLoader, BaseSectionLoaderVars

class AnnouncementLoader(BaseSectionLoader):
    def __init__(self, baseSectionVars: BaseSectionLoaderVars):
        super().__init__(baseSectionVars)

    def load_section(self) -> list[Document]:
        """Load all announcements for a Canvas course"""
        self.logger.logStatement(message='Loading announcements...\n', level="INFO")

        announcement_documents = []
        try:
            announcements = self.canvas_client_extractor.get_announcements()

            for announcement in announcements:
                announcement_documents.extend(self._load_item(announcement=announcement))

        except CanvasException as error:
            self.logger.logStatement(message=f"Canvas exception loading announcements {error}",
                                     level="WARNING")

        return announcement_documents

    def _load_item(self, announcement: DiscussionTopic) -> list[Document]:
        """Loads a single announcement"""
        self.logger.logStatement(message=f"Loading announcement: {announcement.title}", level="DEBUG")
        try:
            announcement_text = self.parse_html(html=announcement.message)
            metadata={"content": announcement_text,
            "data": {"filename": announcement.title,
                    "source": announcement.html_url,
                    "kind": "announcement",
                    "id": announcement.id}
            }
            return self.process_data(metadata=metadata)
        except Exception as error:
            self.logger.logStatement(message=f"Error loading announcement {announcement.title}: {error}",
                                     level="WARNING")
        return []



from typing import Literal
from langchain.document_loaders.base import BaseLoader
from langchain.docstore.document import Document
from pydantic import BaseModel

from canvas_langchain.client import CanvasClient
from canvas_langchain.utils.logging import Logger

# Prevents conflicts with other classes in UMGPT - Happy to refactor as needed
class LogStatement(BaseModel):
    """INFO can be user-facing statements, non-technical and perhaps very high-level"""
    message: str
    level: Literal['INFO', 'DEBUG', 'WARNING']

    def __json__(self):
        return {
            'message': self.message,
            'level': self.level,
        }

class CanvasLoader(BaseLoader):
    def __init__(self,
                 api_url: str,
                 api_key: str,
                 course_id: int, 
                 index_external_urls: bool=False):
        self.logger = Logger()
        self.canvas_client = CanvasClient(api_url, api_key, course_id)
        self.index_external_urls = index_external_urls
        self.course_id = course_id

    def load(self) -> list[Document]:
        """Loads all available content from Canvas course"""
        self.logger.logStatement(message="Starting document loading process. \n", level="INFO")
        docs = []
        try:
            available_tabs = self.canvas_client.get_available_tabs()
            loaders = self.canvas_client.get_loaders(index_external_urls=self.index_external_urls, 
                                                     logger=self.logger)

            for tab_name in available_tabs:
                if tab_name in loaders:
                    docs.extend(loaders[tab_name].load_section())

        except Exception as err:
                self.logger.logStatement(message=f"Error loading Canvas materials {err}", level="WARNING")
        self.logger.logStatement(message="Canvas course processing finished.", level="INFO")
        return docs


    def get_details(self, level='INFO') -> list:
        if level == 'INFO':
            return self.logger._filtered_statements_by_level(level=level)
        return self.logger.progress, self.logger.errors
